/*
	LED banner - An LED Banner Display
	Copyright © 2006-2007 Edward Barnett and Harry Whitfield

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the
	Free Software Foundation; either version 2 of the License, or (at your
	option) any later version.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc.,
	51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

	LED Banner - version 2.2
	15 October, 2007
	Copyright © 2006-2007 Edward Barnett and Harry Whitfield
	mailto:g6auc@arrl.net
*/

var ledWidth  = 14;
var ledHeight = 14;

var charCols = 6;
var charRows = 7;

var bgOffset = 22;	// to allow for RSS QuickList button

var displayLength;
var scale;
var displayMode;
var textMode;
var linearMode;
var theBanner;
var bannerPref;
var filePref;
var theText;
var permitLowerCase;
var frameTime;
var pauseTime;
var colorize;
var frameColor;
var reloadFlag;
var autoStartRSS;
var rssEntry;
var maxRssItems;
var showTitlesOnly;
var showFeedName;
var removeCDATAtags;

var timePref;
var dayPref;
var monthPref;
var datePref;
var yearPref;

var blockLength;

var bannerLength;
var totalLength;

var nBlocks;

var gBlockCount;
var gDisplayCols;
var gBannerCols;
var gTotalCols;

var gScroll;
var gScrollCols;
var gExtras;

var bkgd = null;
var col = [];

function multiClick()
{
	theTimer.ticking = ! theTimer.ticking;
}

function makeDisplay(displayLength, charCols, charRows)
{
	var w = charCols*displayLength;
	var h = charRows + 2;
	bkgd = new Image();
	bkgd.window = main_window;
	bkgd.alignment = 'left';
	bkgd.hOffset = 0;
	bkgd.vOffset = bgOffset;
	bkgd.src = 'Resources/Images/bkgd.png';
	bkgd.opacity = 255;
	bkgd.colorize = frameColor;
	bkgd.zOrder = 1;
	bkgd.onMultiClick = multiClick;
	bkgd.tooltip = "Double-click to pause/continue.";
	bkgd.width  = w*ledWidth*scale;
	bkgd.height = h*ledHeight*scale;
	main_window.width  = w*ledWidth*scale;
	main_window.height = h*ledHeight*scale + bgOffset;
}

function makeMatrix(displayLength, charCols, charRows)
{
	function matrix(displayLength)
	{
		for (var i = 0; i < displayLength; i += 1)
		{
			col[i] = new Image();
			var ch = col[i];
				ch.window = main_window;
				theFrame.addSubview(ch);
				ch.alignment = 'left';
				ch.hOffset = charCols*ledWidth*i*scale;
				ch.vOffset = 0;
				ch.width  = charCols*ledWidth*scale;
				ch.height = charRows*ledHeight*scale;
				ch.src = 'Resources/Images/32.png';
				ch.opacity = 255;
				ch.colorize = colorize;
				ch.zOrder = 3;
		}
	}
	matrix(displayLength);
	theFrame.width   = charCols*ledWidth*displayLength*scale;
	theFrame.height  = charRows*ledHeight*scale;
	theFrame.vOffset = ledHeight*scale + bgOffset;
	theFrame.scrollX = 0;
}
 
function clearBanner(q, n)	//clear q chars at position n
{
	for (var p = 0; p < q; p += 1)
 	{
 		col[p+n].src = 'Resources/Images/32.png';
  		col[p+n].colorize = colorize;
 	}
}
 
var latinMap = [
 32, 33, 99, 35, 35, 89,124, 32, 32, 67, 32, 60,126, 45, 82, 32,
 32, 32, 50, 51, 32,117, 32, 32, 32, 49, 32, 62, 32, 32, 32, 63,
 65, 65, 65, 65, 65, 65, 65, 67, 69, 69, 69, 69, 73, 73, 73, 73,
 68, 78, 79, 79, 79, 79, 79,120, 48, 85, 85, 85, 85, 89, 80,115,
 97, 97, 97, 97, 97, 97, 97, 99,101,101,101,101,105,105,105,105,
100,110,111,111,111,111,111, 47, 48,117,117,117,117,121,112,121];

var kanaMap = [
32,39,49,40,50,41,51,42,52,43,53,54,54,55,55,56,
56,57,57,58,58,59,59,60,60,61,61,62,62,63,63,64,
64,65,65,47,66,66,67,67,68,68,69,70,71,72,73,74,
74,74,75,75,75,76,76,76,77,77,77,78,78,78,79,80,
81,82,83,44,84,45,85,46,86,87,88,89,90,91,39,92,
92,92,38,93,51,54,57,92,92,92,38,37,48,36,36,95];

var kanaMark = [
32,39,49,40,50,41,51,42,52,43,53,54, 0,55, 0,56,
 0,57, 0,58, 0,59, 0,60, 0,61, 0,62, 0,63, 0,64,
 0,65, 0,47,66, 0,67, 0,68, 0,69,70,71,72,73,74,
 0, 1,75, 0, 1,76, 0, 1,77, 0, 1,78, 0, 1,79,80,
81,82,83,44,84,45,85,46,86,87,88,89,90,91,39,92,
92,92,38,93, 0,54,57, 0, 0, 0, 0,37,48,36, 0,95];

function translateBanner(s)
{
	var t = "";
	for (var p = 0; p < s.length; p += 1)
 	{
  		var ch = s.charCodeAt(p);

		if (ch == 0x3000) { ch = 0x0020; } else if (ch == 0x3001) { ch = 0x002C; } else if (ch == 0x3002) { ch = 0x002E; }		// space, full stop and comma
		else if (ch == 0x3008) { ch = 0x003C; } else if (ch == 0x3009) { ch = 0x003E; }											// < and >
  		else if ((0x3041 <= ch) && (ch <= 0x3096)) { ch += 0x60; }	// translate hiragana chars to katakana chars.
  		else if ((0x309D <= ch) && (ch <= 0x309E)) { ch += 0x60; }	// translate hiragana chars to katakana chars.

		t += String.fromCharCode(ch);

 		if ((0x30A0 <= ch) && (ch <= 0x30FF))						// check katakana chars
 		{
 			if (kanaMap[ch - 0x30A0] != kanaMark[ch - 0x30A0]) { t += String.fromCharCode(kanaMark[ch - 0x30A0] + 0x3099); }	// voicing marks
 		}
 	}
 	return t;
}

function loadBanner(s, n)	// load string s at position n
{
	for (var p = 0; p < s.length; p += 1)
 	{
 		var ch = s.charCodeAt(p);
  		if (! permitLowerCase)
  		{
  			if ((97 <= ch) && (ch <= 122)) { ch = ch - 32; }
  			else if ((0x00E0 <= ch) && (ch <= 0x00FE) && (ch != 0x00F7)) { ch = ch - 32; }
  		}
  		var path = 'Resources/Images/' + ch + '.png';
  		if (ch < 32) { col[p+n].src = 'Resources/Images/32.png'; }
 		else if ((32 <= ch) && (ch <= 127)) { col[p+n].src = path; }
 		else if ((0x00A0 <= ch) && (ch <= 0x00FF))  { col[p+n].src = path; }
/*
 		{
  			if (filesystem.itemExists(path)) { col[p+n].src = path; }
 			else
 			{
 				ch = latinMap[ch - 0x00A0];								// map missing accented characters
 				col[p+n].src = 'Resources/Images/' + ch + '.png';
 			}
 		}
*/
 		else if (ch == 0x20AC) { col[p+n].src = 'Resources/Images/' + ch + '.png'; }
 		else if ((0x30A0 <= ch) && (ch <= 0x30FF)) { col[p+n].src = 'Resources/Katakana/' + (kanaMap[ch - 0x30A0] + 128) + '.png'; }
  		else if ((0x31F0 <= ch) && (ch <= 0x31FF)) { col[p+n].src = 'Resources/Katakana/' + (ch - 0x31F0 + 224) + '.png'; }
  		else if ((0x3099 <= ch) && (ch <= 0x309A)) { col[p+n].src = 'Resources/Katakana/' + (ch - 0x3099 + 153) + '.png'; }
  		else if ((0x309B <= ch) && (ch <= 0x309C)) { col[p+n].src = 'Resources/Katakana/' + (ch - 0x309B + 153) + '.png'; }
  		else if  (ch == 0x309F) { col[p+n].src = 'Resources/Katakana/159.png'; }
 		else { col[p+n].src = 'Resources/Images/127.png'; }
  		col[p+n].colorize = colorize;
 	}
}
 
function block(i) { return theText.substr(i*blockLength, blockLength); }

function updateImage()
{
 	gScroll = (gScroll + 1); theFrame.scrollX = -gScroll*ledWidth*scale;
  	if (gScroll == gScrollCols)
  	{
  		updateNow();
  		
 		if (gBlockCount === 0)
		{
			if (textMode) { if (gExtras) { theText = getPaddedBanner(bannerPref); } }
			else if (reloadFlag) { getDisplayText(true); }									// was { reloadWidget(); }
			if (pauseTime != "0") { theTimer.interval = pauseTime; }
		}

		if (linearMode)
 		{
				if (((gBlockCount + 1) % nBlocks) === 0) { loadBanner(block(gBlockCount), 0); clearBanner(blockLength, blockLength); }
				else { loadBanner(block(gBlockCount) + block(gBlockCount + 1), 0); }
 		}
 		else
 		{
				loadBanner(block(gBlockCount) + block((gBlockCount + 1) % nBlocks), 0);
  		}
  		gBlockCount = (gBlockCount + 1) % nBlocks;
  		gScroll = 0;
  		theFrame.scrollX = 0;
  	} 
 	else { theTimer.interval = frameTime; }
 	
}
 
function reloadRSS()
{
	main_window.visible = false;
	updateNow();
	if (linearMode) { preferences.modePref.value = "Linear RSS"; } else { preferences.modePref.value = "Cyclic RSS"; }
	ledPrefs();
	//reloadFlag = false;
 	getDisplayText(false);

	gScroll = 0;
	gBlockCount = 1;
	theFrame.scrollX = 0;
	loadBanner(block(0) + block(1), 0);
	showButtons();
	main_window.visible = true;
	updateNow();
	
	if (pauseTime != "0") { theTimer.interval = pauseTime; } else { theTimer.interval = frameTime; }
	theTimer.ticking = true;
}

function getTime(hour24, AMPM)
{
	var d = new Date(); var h = d.getHours(); var m = d.getMinutes(); var ampm = "";
	h = h % 24;   if ( m < 10 ) { m = "0" + m; }
	if (hour24) { if ( h < 10 ) { h = "0" + h; } }
	else
	{
		ampm = "AM"; if (h >= 12) { ampm = "PM"; if (h > 12) { h = h - 12; } }
		if ( h === 0 ) { h = 12; }
		if ( h < 10 ) { h = " " + h; }
		if ( ! AMPM  ) { ampm = ""; }
	}
	return h.toString() + ':' + m.toString() + ampm;
}

function getDate(day, month, date, year)
{
	var theDays   = ["Sun","Mon","Tue","Wed","Thu","Fri","Sat"];
	var theMonths = ["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"];

	var d = new Date();
	var theDay   = theDays[d.getDay()];
	var theMonth = theMonths[d.getMonth()];
	var theDate  = d.getDate();	if (theDate < 10) { theDate = ' ' + theDate; }
	var theYear  = d.getFullYear();
	
	var result = "";
	if (day)   { result += theDay   + ' '; }
	if (month) { result += theMonth + ' '; }
	if (date)  { result += theDate  + ' '; }
	if (year)  { result += theYear  + ' '; }
	
	return result;
}

function getBanner(banner)
{
	// add extra texts in here - such as time and date
	
	gExtras = false;

	var hour24 = false; var AMPM = true;
	switch(timePref)
	{
		case "12-hour with AM/PM": break;
		case "12-hour": AMPM = false; break;
		case "24-hour": hour24 = true; break;
		default: break;
	}
	if (timePref !== "none") { banner = getTime(hour24, AMPM)     + '  ' + banner; gExtras = true; }
	
	var displayDate = dayPref || monthPref || datePref || yearPref;

	if (displayDate) { banner = getDate(dayPref, monthPref, datePref, yearPref) + ' ' + banner; gExtras = true; }

	return banner;
}

function getPaddedBanner(banner)
{
	var theText = getBanner(banner);
	while ((theText.length % blockLength) !== 0) { theText += ' '; }
	if (theText.length == blockLength) { theText += theText; }
	return theText;
}

function escapePath(s) { return s.replace(/([\W])/g,"\\$1"); }

function xtn(s)
{
    var idx = s.lastIndexOf(".");
    if (idx >= 0)
    {
        var ext = s.substring(idx);
        if (ext.length > 7) { return  ""; }
        if (ext.length > 1) { return ext; }
    }
   	return "";
}

function isItemIn(item, vector)
{
	for (var i = 0; i < vector.length; i += 1)
	{
		if (vector[i] == item) { return true; }
	}
	return false;
}

function isText(path)
{
	var textExtns = [".txt",".js",".kon",".c",".bat",".css",".csv",".dtd",".faq",".jav",".js",".log",".pl",".sig",".xml",".htm","html"];

	if (systemPlatform == "macintosh")
	{
		var result = getFileInfo(path);
		var fileData = result.substr(path.length + 2);
		// print('isText:Mac fileType=' + fileData);
		var item = fileData.split(" ");
		var stype = item[0];
		if ((stype == "ASCII") || (stype == "ISO-8859") || (stype == "UTF-8")) { return true; }
	}
	var xtype = xtn(path).toLowerCase();
	// print('isText:Win fileType=' + xtype);
	return isItemIn(xtype, textExtns);
}

function openRSSdownloader()
{
	filesystem.open('Resources/RSS Downloader.widget');
}

function checkFile(path)
{
	return (filesystem.itemExists(path) && (! filesystem.isDirectory(path)));
}

function ledPrefs()
{
	// pick up preferences
	
	displayLength   = Number(preferences.sizePref.value);
	scale			= preferences.heightPref.value / 100;
	displayMode     = preferences.modePref.value;
	filePref		= preferences.filePref.value;
	permitLowerCase = (preferences.lcPref.value == "1");
	frameTime       = 1 / Number(preferences.timerPref.value);
	pauseTime       = preferences.pausePref.value;
	colorize        = preferences.colorPref.value;
	frameColor      = preferences.bgColorPref.value;
	reloadTimerPref = preferences.reloadTimerPref.value;
	bannerPref      = preferences.bannerPref.value;
	autoStartRSS    = (preferences.autoStartRSS.value == "1");
	rssEntry 		= preferences.savedDefaultPref.value;
	maxRssItems	    = Number(preferences.rssItemsPref.value);
	showTitlesOnly  = (preferences.rssTitlePref.value == "1");
	showFeedName    = (preferences.rssFeedPref.value  == "1");
	removeCDATAtags = (preferences.rssCDATAPref.value == "1");

	timePref        = preferences.timePref.value;
	dayPref         = (preferences.dayPref.value   == "1");
	monthPref       = (preferences.monthPref.value == "1");
	datePref        = (preferences.datePref.value  == "1");
	yearPref        = (preferences.yearPref.value  == "1"); 
	
	blockLength = 2*displayLength;
	
	textMode   = (displayMode.indexOf("Text")   >= 0);
	fileMode   = (displayMode.indexOf("File")   >= 0);
	rssMode    = (displayMode.indexOf("RSS")    >= 0);
	linearMode = (displayMode.indexOf("Linear") >= 0);
	
	gDisplayCols    = displayLength*charCols;
}
	
function startDownloader()
{

	if (autoStartRSS)
	{
		var oldFileExists = checkFile(filePref);
		
		openRSSdownloader();
		
		if (fileMode)
		{
			if (oldFileExists) { sleep(5000); }			// give the RSS Downloader 5 seconds to replace the old file
	
			for (var count = 0; count < 10; count += 1)	// give the RSS Downloader up to 10 seconds to fetch a file
			{
				if (checkFile(filePref)) { break; }
				sleep(1000);
			}
		}
	}
}
	
function getDisplayText(isReload)
{
	if (rssMode)
	{
		if (isReload) { cycleRSSLocation(); rssEntry = preferences.savedDefaultPref.value; }
		reloadFlag = false;
		if (reloadTimerPref != "0")
		{
			theReloadTimer.interval = reloadTimerPref;
			theReloadTimer.ticking = true;
		}
		theText = getRSSdata(rssEntry);
		var ok = !!theText;
		if (! ok) { beep(); alert('Could not get RSS data. Please choose another mode.'); showWidgetPreferences(); closeWidget(); }
		var rssFeedName = extractName(rssEntry);
		if (showFeedName) { theText = rssFeedName + ' :: ' + theText; }
		theText = translateBanner(theText);
		quickMenu.tooltip = "Displaying " + rssFeedName;
	}
	else if (fileMode)
	{
		reloadFlag = false;
		if (reloadTimerPref != "0")
		{
			theReloadTimer.interval = reloadTimerPref;
			theReloadTimer.ticking = true;
		}
		ok = isText(filePref);
		if (ok)
		{
			theText = filesystem.readFile(filePref);
			ok = !!theText;
		}
		if (! ok) { beep(); alert('Please select a text file or choose another mode.'); showWidgetPreferences(); closeWidget(); }
		theText = translateBanner(theText);
		quickMenu.tooltip = "Displaying a file.";
	}
	else
	{
		theText = bannerPref = translateBanner(bannerPref);
		quickMenu.tooltip = "Displaying a text.";
	}
	theText = getPaddedBanner(theText);
	nBlocks = theText.length / blockLength;
}

function createDisplay()
{
	gScroll = 0;
	gBlockCount = 1;
	bannerLength = blockLength;
	totalLength  = 2*bannerLength;
	gScrollCols  = gBannerCols = bannerLength*charCols;
	gTotalCols   = totalLength*charCols;
	makeDisplay(displayLength, charCols, charRows);
	makeMatrix(totalLength,   charCols, charRows);
	loadBanner(block(0) + block(1), 0);
	if (pauseTime != "0") { theTimer.interval = pauseTime; } else { theTimer.interval = frameTime; }
	theTimer.ticking = true;
}

function ledDisplay()
{
	ledPrefs();
	//startDownloader();
	getDisplayText(false);
	createDisplay();
}
